#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include "lsm/store.h"
#include "lsm/store_internal.h"

static lsm_error lsm_entry_write_uint64_t(FILE *f, uint64_t num) {
  size_t res = fwrite(&num, sizeof(uint64_t), 1, f);

  // Such a small write should succeed in one go
  if (res == 0) {
    return lsm_error_failed_io;
  }

  return lsm_error_ok;
}

static lsm_error lsm_entry_write_str(FILE *f, lsm_str *s) {
  uint64_t to_write = lsm_str_len(s);
  uint64_t written = 0;

  do {
    written += fwrite(lsm_str_ptr(s), sizeof(char), to_write - written, f);
  } while (written < to_write);

  return lsm_error_ok;
}

lsm_error lsm_entry_write_db(uint64_t *size, FILE *db_file, lsm_entry *entry) {
  // First we write how many attributes follow
  LSM_RES(lsm_entry_write_uint64_t(db_file, entry->attrs.count));
  *size = sizeof(uint64_t);

  for (uint64_t i = 0; i < entry->attrs.count; i++) {
    // Write attribute type, length & value
    LSM_RES(lsm_entry_write_uint64_t(db_file, entry->attrs.items[i].type));
    LSM_RES(lsm_entry_write_uint64_t(db_file, lsm_str_len(entry->attrs.items[i].str)));
    LSM_RES(lsm_entry_write_str(db_file, entry->attrs.items[i].str));

    *size += 2 * sizeof(uint64_t) + lsm_str_len(entry->attrs.items[i].str) * sizeof(char);
  }

  return lsm_error_ok;
}

lsm_error lsm_entry_write_idx(uint64_t *size, FILE *idx_file, lsm_entry *entry, uint64_t offset, uint64_t len) {
  LSM_RES(lsm_entry_write_uint64_t(idx_file, lsm_str_len(entry->key)));
  LSM_RES(lsm_entry_write_str(idx_file, entry->key));
  LSM_RES(lsm_entry_write_uint64_t(idx_file, offset));
  LSM_RES(lsm_entry_write_uint64_t(idx_file, len));

  *size = 3 * sizeof(uint64_t) + lsm_str_len(entry->key) * sizeof(char);

  return lsm_error_ok;
}

lsm_error lsm_entry_sync(lsm_store *store, lsm_entry_handle *handle) {
  pthread_mutex_lock(&store->db_lock);

  // Append entry to end of database file
  if (fseek(store->db_file, store->db_file_size, SEEK_SET) != 0) {
    pthread_mutex_unlock(&store->db_lock);

    return lsm_error_failed_io;
  }

  uint64_t entry_size;
  lsm_error res = lsm_entry_write_db(&entry_size, store->db_file, handle->wrapper->entry);
  fflush(store->db_file);

  // TODO fsync db file?

  if (res != lsm_error_ok) {
    pthread_mutex_unlock(&store->db_lock);

    return res;
  }
  
  uint64_t entry_index = store->db_file_size;
  store->db_file_size += entry_size;

  pthread_mutex_unlock(&store->db_lock);

  // Append entry to index file
  pthread_mutex_lock(&store->idx_lock);

  if (fseek(store->idx_file, store->idx_file_size, SEEK_SET) != 0) {
    printf("failed seek, %lu\n", store->idx_file_size);
    pthread_mutex_unlock(&store->idx_lock);

    return lsm_error_failed_io;
  }

  res = lsm_entry_write_idx(&entry_size, store->idx_file, handle->wrapper->entry, entry_index, entry_size);

  if (res == lsm_error_ok) {
    // Update the counter at the beginning of the file
    uint64_t new_block_count = store->idx_file_block_count + 1;

    if (fseek(store->idx_file, 0, SEEK_SET) != 0) {
      pthread_mutex_unlock(&store->idx_lock);

      return lsm_error_failed_io;
    }

    size_t r = fwrite(&new_block_count, sizeof(uint64_t), 1, store->idx_file);

    if (r != lsm_error_ok) {
      printf("wuck\n");
      pthread_mutex_unlock(&store->idx_lock);

      return res;
    }

    store->idx_file_size += entry_size;
    store->idx_file_block_count = new_block_count;
  } else {
    printf("failed write\n");
  }

  fflush(store->idx_file);

  pthread_mutex_unlock(&store->idx_lock);

  return res;
}

static lsm_error lsm_entry_read_attrs(lsm_entry_handle *handle, FILE *db_file) {
  uint64_t attr_count;
  size_t res = fread(&attr_count, sizeof(uint64_t), 1, db_file);

  if (res == 0) {
    return lsm_error_failed_io;
  }

  // attr_type, val_len
  uint64_t nums[2];
  lsm_str *val;

  for (uint64_t i = 0; i < attr_count; i++) {
    res = fread(nums, sizeof(uint64_t), 2, db_file);

    if (res < 2) {
      return lsm_error_failed_io;
    }

    char *val_s = malloc(nums[1] + 1);
    val_s[nums[1]] = '\0';

    if (val_s == NULL) {
      return lsm_error_failed_alloc;
    }
    
    uint64_t read = 0;

    while (read < nums[1]) {
      read += fread(&val_s[read], 1, nums[1] - read, db_file);
    }

    LSM_RES(lsm_str_init(&val, val_s));;
    lsm_entry_attr_insert(handle, nums[0], val);
  }

  return lsm_error_ok;
}

lsm_error lsm_store_load_db(lsm_store *store) {
  uint64_t key_len;
  uint64_t db_dim[2];
  lsm_str *key;
  lsm_entry_handle *handle;

  // idx file starts with block count
  size_t res = fread(&store->idx_file_block_count, sizeof(uint64_t), 1, store->idx_file);

  if (res == 0) {
    return lsm_error_failed_io;
  }

  store->idx_file_size += sizeof(uint64_t);

  for (uint64_t i = 0; i < store->idx_file_block_count; i++) {
    // Read in idx metadata
    res = fread(&key_len, sizeof(uint64_t), 1, store->idx_file);

    if (res == 0) {
      return lsm_error_failed_io;
    }

    char *key_s = malloc(key_len + 1);
    key_s[key_len] = '\0';

    if (key_s == NULL) {
      return lsm_error_failed_alloc;
    }

    res = fread(key_s, 1, key_len, store->idx_file);

    if (res < key_len) {
      return lsm_error_failed_io;
    }

    res = fread(db_dim, sizeof(uint64_t), 2, store->idx_file);

    if (res < 2) {
      return lsm_error_failed_io;
    }

    LSM_RES(lsm_str_init(&key, key_s));
    LSM_RES(lsm_store_insert(&handle, store, key));

    // Read attributes from database file
    if (fseek(store->db_file, db_dim[0], SEEK_SET) != 0) {
      return lsm_error_failed_io;
    }

    LSM_RES(lsm_entry_read_attrs(handle, store->db_file));
    lsm_entry_close(handle);

    store->idx_file_size += 3 * sizeof(uint64_t) + key_len;
    store->db_file_size += db_dim[1];
  }

  return lsm_error_ok;
}
